"""Functions and classes for a display that allows changing the information about a thing."""

import asyncio
import itertools
import os
import urllib.request
from typing import (
    Union,
    Callable,
    Any,
    Literal,
    Optional,
    override,
    cast,
    Iterable,
    MutableMapping,
)
import urllib.error
import i18n

import nicegui
from nicegui import ui

import flinventory
from flinventory import BoxedThing, Sign


def t(key: str, **kwargs) -> str:
    """Translate a string, suitable for this module and the current user locale.

    Translation string for this module are located in translation/search.LOCALE.yml.

    Fallback locale is "en" but this should never be used because the locale should
    be set to the primary language of the inventory.

    Args:
        key: the key in the yml file for the string looked for
        kwargs: further arguments for the i18n.t function
    """
    return i18n.t(
        f"thing_editor.{key}",
        locale=nicegui.app.storage.user.get("locale", "en"),
        **kwargs,
    )


def try_conversion(
    value: str, conversions: Iterable[Union[Callable[[str], Any], Literal["bool"]]]
) -> Any:
    """Try to convert value into other values.

    Args:
        value: value to be converted
        conversions: tuple of conversion functions that raise Exception upon failure
            or "bool" which converts the strings "true", "false"
    """
    for conversion in conversions:
        if conversion == "bool":
            try:
                low_value = value.lower()
            except AttributeError:
                continue
            if low_value == "true":
                return True
            if low_value == "false":
                return False
            continue
        try:
            return conversion(value)
        except Exception:  # pylint: disable=broad-exception-caught
            pass
    return value


class SlowInputSaver[Container: MutableMapping]:
    """Saves content of an input field to a dict-like (a BoxedThing).

    Does so with a bit of delay to avoid creating save operations if
    new letters are typed immediately after.

    Can be used as an on-change-action connected to an input field.

    Maybe it makes sense to make SlowInputSaver dependent on the type of the data
    object. We will see.
    """

    def __init__(
        self,
        data: Container,
        member: flinventory.defaulted_data.Key,
        wait_time: Union[float, int] = 1,
    ):
        """Create a SlowInputSaver.

        Args:
            data: the data object (think thing) to be altered
            member: which information to be altered. thing[member] = <something suitable>
                must be allowed
            wait_time: how long (in seconds) the saver should wait
                until it assumes no further changes and actually saves input
        """
        self.data = data
        self.member = member
        self.wait_time = wait_time
        self._running_queries: list[asyncio.Task] = []

    def converter(self, value: Any) -> Any:
        """A hook for subclasses to override for value conversion.

        Since subclasses might do different things with it, it needs to stay a method.
        The linter gets it by actually overriding it in another class. (Where it still
        could be a static method.)
        """
        return value

    def setter(self, value: Any):
        """How the value of the input element is saved.

        This can be overridden by subclasses.
        """
        if not value:
            try:
                del self.data[self.member]
            except KeyError:
                # already gone, so already fine
                pass
        else:
            self.data[self.member] = value

    async def __call__(self, event: nicegui.events.ValueChangeEventArguments):
        """Value has changed. Now save it. If saving is not cancelled by further changes."""
        for query in self._running_queries:
            query.cancel()
        sleep = asyncio.create_task(asyncio.sleep(self.wait_time))
        self._running_queries.append(sleep)
        try:
            await sleep
        except asyncio.exceptions.CancelledError:
            # the next letter was already typed, do not search and rerender for this query
            return
        self.setter(self.converter(event.value))


class SlowSignChanger(SlowInputSaver[flinventory.Sign]):
    """Special case for signs: add resetting printed attribute."""

    @override
    def setter(self, value: Any):
        """Sets the value and sets printed to false."""
        if not value:
            if "printed" in self.data and self.member in self.data:
                del self.data["printed"]
            try:
                del self.data[self.member]
            except KeyError:
                # already not there, fine
                pass
        else:
            previous = self.data.get(self.member, "")
            self.data[self.member] = try_conversion(
                value,
                (
                    cast(Callable[[str], Any], int),
                    cast(Callable[[str], Any], float),
                    "bool",
                ),
            )
            if previous != self.data[self.member]:
                try:
                    del self.data["printed"]
                except KeyError:
                    pass


class SlowListChanger(SlowInputSaver):
    """Special case for alt names: add splitting at ;

    This is for simple lists with several strings in one input field.
    """

    @override
    def converter(self, value: Any) -> Any:
        """Split at ;."""
        return list(filter(None, map(str.strip, value.split(";"))))


def update_location_element(
    location_ui_element, loc: flinventory.Location, focus: Optional[str] = None
):
    """List location information with input fields.

    Args:
        location_ui_element: ui_element which content get replaced
        loc: location data to be displayed
        focus: the level name of the schema which input field should have the cursor
            None if no focus should be set
    """
    location_ui_element.clear()
    with location_ui_element:
        location_info = loc.to_jsonable_data()
        schema_hierarchy = loc.schema.get_schema_hierarchy(loc)
        ui.label(str(schema_hierarchy[0].name))
        editable_levelnames = []
        for level in schema_hierarchy:
            try:
                level_name = level.levelname
            except flinventory.location.InvalidLocation:
                # bottom most location (in hierarchy) has no subs and therefore
                # needs no input
                continue
            else:
                editable_levelnames.append(level_name)
            # some Value error because it gets value Werkstatt from somewhere:
            if subs := level.get_valid_subs(shortcuts="no"):
                current_value = location_info.get(level.levelname, "")
                options = (
                    ([current_value] if current_value not in subs else [])
                    + ([""] if (current_value != "" and "" not in subs) else [])
                    + subs
                )
                ui.select(
                    options=options,
                    label=t(level.levelname),
                    value=current_value,
                    clearable=True,
                    with_input=True,
                    # new_value_mode="add",
                    on_change=SlowLocationSaver(
                        loc, level, location_ui_element, wait_time=0
                    ),
                ).props(
                    "autogrow dense" + (" autofocus" if level_name == focus else "")
                )
            else:
                ui.input(
                    label=t(level.levelname),
                    value=str(location_info.get(level.levelname, "")),
                ).props(
                    "autogrow dense" + (" autofocus" if level_name == focus else "")
                ).on_value_change(
                    SlowLocationSaver(loc, level, location_ui_element)
                )
        additional_info = [
            (key, value)
            for key, value in location_info.items()
            if not key in editable_levelnames
        ]
        if additional_info:
            ui.label(t("unused_location_info") + ":")
        for key, value in additional_info:
            ui.label(f"{t(key)}: {value}")


def update_sign_element(
    sign_ui_element: ui.element, sign: Sign, options: flinventory.Options
):
    """Replace content of sign_ui_element with input fields for the sign.

    Args:
        sign_ui_element: a nice_gui element, usually an ui.column which content
            gets replaced by input fields
        sign: the sign that should be displayed and is changable by the new input fields
        options: options including the languages
    """
    input_fields: dict[tuple[str, str] | str, ui.element] = {}
    with sign_ui_element:
        for sign_member, lang, tooltip in (
            ("width", "", "[cm]"),
            # cm needs to be replaced by options.length_unit once this option is actually used
            ("height", "", "[cm]"),
            ("name", options.languages[0], t("default_main_name")),
            ("name", options.languages[1], t("default_main_name")),
            (
                "fontsize",
                options.languages[0],
                f"{t("default")}: {flinventory.signprinter_latex.STANDARD_FONT_SIZE_GERMAN}",
            ),
            (
                "fontsize",
                options.languages[1],
                f"{t("default")}: {flinventory.signprinter_latex.STANDARD_FONTSIZE_ENGLISH}",
            ),
            (
                "location_shift_down",
                "",
                f"{t("default")}: {flinventory.signprinter_latex.STANDARD_LOCATION_SHIFT_DOWN}",
            ),
        ):
            label = f"{t('sign')} {t(sign_member)} {f'({t(lang)})' if lang else ''}"
            member = (sign_member, lang) if lang else sign_member
            input_fields[f"sign.{member}"] = (
                ui.input(
                    label=label,
                    value=str(sign.get_undefaulted(member, backup="")),
                )
                .props("autogrow dense")
                .on_value_change(SlowSignChanger(sign, member))
            )
            if tooltip:
                input_fields[f"sign.{member}"].tooltip(tooltip)


class SlowImageFetcher(SlowInputSaver[flinventory.Thing]):
    """Special case for image url.

    In addition to saving the inputted URL,
    this saver tries to fetch the specified url and saves it as an image if it is one.

    If the fetched file is huge, this might be problematic.
    Do I assume no malicious actor?
    """

    @override
    def converter(self, value: str) -> str:
        """Add http if no protocol is given."""
        if "://" not in value:
            return "http://" + value
        return value

    @override
    def setter(self, value: str):
        """Saves the URL and tries to save the linked image.

        If the url is removed, the image is not removed. Need a separate button for that.
        If an image exists, it is overwritten.
        It is assumed that there is some data backup system like a git repository.

        todo: show image after download.
        """
        super().setter(value)
        try:
            if any(
                value.lower().endswith(extension := image_type)
                for image_type in flinventory.constant.IMAGE_FILE_TYPES
            ):
                extension = "." + extension
            else:
                extension = ""
            urllib.request.urlretrieve(
                value,
                os.path.join(
                    self.data.directory,
                    flinventory.constant.IMAGE_FILE_BASE + extension,
                ),
            )
        except urllib.error.ContentTooShortError as interrupted:
            ui.notification(
                f"Error downloading image from {value}: \n"
                f"reason: {interrupted.reason}\n"
                f"aim filename: {interrupted.filename}\n"
                f"other info: {interrupted.args}, {interrupted}"
            )
        except ValueError as unknown_url:
            ui.notification(f"No download of {value} possible: {unknown_url}")
        except urllib.error.HTTPError as http_error:
            ui.notification(f"Download unsuccessful: {http_error}.")
        except FileNotFoundError as directory_missing:
            ui.notification(
                f"Saving image failed. Programming error: {directory_missing}"
            )


class SlowLocationSaver(SlowInputSaver[flinventory.Location]):
    """Special case for location info."""

    def __init__(
        self,
        data: flinventory.Location,
        level: flinventory.Schema,
        ui_element: ui.element,
        wait_time: float = 0.3,
    ):
        """Create a special saver that handles location data.

        Args:
            data: the location
            level: the level including the level name that is changed
            ui_element: the ui element that needs update after a change
            wait_time: time in seconds waited before saving for case of quickly changed input
        """
        # wait less long because I want to quickly get the autocompletion
        # and the next input field
        super().__init__(data, level.levelname, wait_time=wait_time)
        self.level = level
        self.ui_element = ui_element

    @override
    def converter(self, value: Any) -> Any:
        """Convert to bool or int if possible."""
        return try_conversion(value, (int, "bool"))

    @override
    def setter(self, value: Any):
        """Set the value and update the ui."""
        if value in self.data.EMPTY:
            self.data[self.level.levelname] = value
            update_location_element(
                self.ui_element,
                loc=self.data,
                focus=self.level.levelname,
            )
            return
        try:
            sub_schema = self.level.get_subschema(value)
        except flinventory.location.InvalidLocationSchema as error:
            print("InvalidLocationSchema: ", error)
        except flinventory.location.InvalidLocation:
            print(f"InvalidLocation {value}: Do not save")
        else:
            # if possible focus on subschema level because that is the one
            # that most likely needs to be edited next:
            try:
                focus = sub_schema.levelname
            except flinventory.location.InvalidLocation:
                focus = self.level.levelname
            self.data[self.level.levelname] = value
            update_location_element(
                self.ui_element,
                loc=self.data,
                focus=focus,
            )


class SlowThingListAdder(SlowInputSaver):
    """Adds a thing id to a list if input is unique.

    This is for lists where the elements are thing IDs.
    """

    def __init__(
        self,
        inventory: flinventory.inventory_io.Inventory,
        thing: flinventory.defaulted_data.DefaultedDict,
        member: flinventory.defaulted_data.Key,
        list_ui_element: ui.element,
        input_element: ui.input,
        description: Optional[str],
    ):
        """Create the ListRemover and tell it what to remove and what to update afterward.

        Args:
            inventory: Inventory needed for calling update_list_input
            thing: defaulted dict (so far only things) to be changed
            member: which list should be changed (the key)
            value_to_be_removed: the value to be removed (not its index)
            list_ui_element: nicegui ui element to be called update_list_input
            input_element: nicegui input element which might get additional letters
            description: argument of update_list_input called afterward
        """
        super().__init__(thing, member)
        self.inventory = inventory
        self.list_ui_element = list_ui_element
        self.input_element = input_element
        self.description = description

    @override
    def setter(self, value: Any):
        """Add new thing if value is unique.

        If new thing is added, reload list ui element.
        """
        try:
            new_thing = self.inventory.get_by_id(value)
        except KeyError:
            new_thing_options = itertools.chain(
                self.inventory.get_by_key(("name", lang), value)
                for lang in self.inventory.options.languages
            )
            new_thing = None  # default if Iterator is empty
            for new_thing in new_thing_options:
                # use first element
                break
            # else:
            #     could add code to fill the input field as much as all options agree
            for _ in new_thing_options:
                ui.notify(t("ambiguous_name", name=value))
                new_thing = None
                break
        if new_thing is not None:
            assert isinstance(self.data.get(self.member, tuple()), tuple), (
                f"{self.data.best('name', backup="?")} is not a tuple but "
                f"{self.data.get[self.member]} of type {type(self.data.get[self.member])}"
                f" (in Adder)."
            )
            self.data[self.member] = list(self.data.get(self.member, tuple())) + [
                self.inventory.get_id(new_thing)
            ]
            update_list_input(
                self.list_ui_element,
                self.inventory,
                self.data,
                self.member,
                self.description,
            )


class ListRemover:
    """A function (implemented as a class with __call__) that removes a value from a list.

    Not slow because it is intended for a button which should act immediately.
    """

    def __init__(
        self,
        inventory: flinventory.inventory_io.Inventory,
        thing: flinventory.defaulted_data.DefaultedDict,
        member: flinventory.defaulted_data.Key,
        value_to_be_removed: flinventory.defaulted_data.Immutable,
        ui_element: ui.element,
        description: Optional[str],
    ):
        """Create the ListRemover and tell it what to remove and what to update afterward.

        Args:
            inventory: Inventory needed for calling update_list_input
            thing: defaulted dict (so far only things) to be changed
            member: which list should be changed (the key)
            value_to_be_removed: the value to be removed (not its index)
            ui_element: nicegui ui element to be called update_list_input
            description: argument of update_list_input called afterward
        """
        self.inventory = inventory
        self.thing = thing
        self.member = member
        self.value = value_to_be_removed
        self.ui_element = ui_element
        self.description = description

    def __call__(self, _: Any):
        """Remove the specified element from the list and call update_list_input.

        Ignore the argument.
        """
        current = self.thing.get_undefaulted(self.member, backup=tuple())
        assert isinstance(current, tuple), (
            f"{self.thing.best('name', backup="")}.{self.member} is "
            f"not a tuple but {self.thing[self.member]}."
        )
        new = [value for value in current if value != self.value]
        if new:
            self.thing[self.member] = new
        else:
            del self.thing[self.member]
        update_list_input(
            self.ui_element, self.inventory, self.thing, self.member, self.description
        )


def update_list_input(
    ui_element: ui.element,
    inventory: flinventory.inventory_io.Inventory,
    thing: flinventory.DefaultedDict,
    member: flinventory.defaulted_data.Key,
    description: Optional[str] = None,
) -> None:
    """A set of ui elements for adding and removing references to other things.

    thing[member] must be a tuple. If member is a translatable list member.

    Args:
        ui_element: nicegui element which is cleared and refilled
        inventory: inventory needed to translate names and UUIDs to things
        thing: thing (or other defaulted dict) that should be changed
        member: the key whose value should be changed. Current value can be non-existing.
        description: what is written in front of list to tell the user what s:he is editing.
            If None (default) use member or "_".join(member)
    """
    ui_element.clear()
    if description is None:
        description = member if isinstance(member, str) else "_".join(member)
    with ui_element:
        ui.label(description + ":")
        assert isinstance(thing.get(member, tuple()), tuple), (
            f"{thing.best("name", backup="")}.{member} is not a tuple but "
            f"{thing[member]} of type {type(thing[member])}!"
        )
        for thing_id in thing.get_undefaulted(member, backup=tuple()):
            try:
                other_thing = inventory.get_by_id(thing_id)
            except KeyError:
                # ignore
                continue
            with ui.label():
                ui.link(
                    text=other_thing.best("name", backup="?"),
                    target=f"/thing/{thing_id}",
                    new_tab=True,
                )
                ui.button("X").on_click(
                    ListRemover(
                        inventory, thing, member, thing_id, ui_element, description
                    )
                )
        input_element = ui.input(label=t("new")).props("autogrow dense")
        input_element.on_value_change(
            SlowThingListAdder(
                inventory, thing, member, ui_element, input_element, description
            )
        )


def update_thing_basics_element(
    ui_element: nicegui.ui.element,
    thing: BoxedThing,
    inventory: flinventory.inventory_io.Inventory,
) -> None:
    """Clear content of ui element and instead display editing fields for basic info, like name."""
    ui_element.clear()
    with ui_element:
        for language in inventory.options.languages:
            member: tuple[str, str] | str
            for member in ("name", "description"):
                input_field = (
                    ui.input(
                        label=f"{t(member)} ({language})",
                        value=thing.thing.get_undefaulted(
                            (member, language), backup=""
                        ),
                    )
                    .props("autogrow dense")
                    .on_value_change(SlowInputSaver(thing.thing, (member, language)))
                )  # thing instead of thing.thing works as well but gives type error
                # since thing is not officially a mapping
                if member == "description":
                    input_field.bind_visibility_from(
                        nicegui.app.storage.user, "expert_mode"
                    )
            member = "name_alt"
            ui.input(
                label=f"{t(member)} ({language}) ({t("separated", sep=";")})",
                value="; ".join(
                    thing.thing.get_undefaulted((member, language), backup=[])
                ),
            ).props("autogrow dense").on_value_change(
                SlowListChanger(thing.thing, (member, language))
            )
        ui.input(
            label=t("image_url"),
            value=thing.thing.get_undefaulted("image_url", backup=""),
        ).props("dense").on_value_change(
            SlowImageFetcher(thing.thing, "image_url")
        ).bind_visibility_from(
            nicegui.app.storage.user, "expert_mode"
        )
        for member in ("subclass_of", "part_of"):
            with ui.column().bind_visibility_from(
                nicegui.app.storage.user, "expert_mode"
            ) as super_row:
                update_list_input(
                    super_row,
                    inventory,
                    thing.thing,
                    member,
                    t(member),
                )


def show_thing_changer(
    ui_element: nicegui.ui.element,
    thing: BoxedThing,
    inventory: flinventory.inventory_io.Inventory,
) -> None:
    """Clear content of ui element and instead display editing fields.

    Args:
        ui_element: the ui element (e.g. a card) on which to show the thing changing ui
        inventory: the inventory needed for references to other things. Also used for its options,
            including the languages.
        thing: the thing to change
    """
    options = inventory.options

    print(
        f"Try to let edit {thing.best('name', backup="a new thing")} with {id(ui_element)}."
    )
    ui_element.clear()
    with ui_element:
        with ui.row():
            with ui.column() as basics_column:
                update_thing_basics_element(basics_column, thing, inventory)
            with ui.column() as location_column:
                update_location_element(location_column, thing.location)
            with ui.column() as sign_column:
                sign_column.bind_visibility_from(
                    nicegui.app.storage.user, "expert_mode"
                )
                update_sign_element(sign_column, thing.sign, options)
            if image := thing.thing.image_path():
                ui.image(image).props("width=50%").props("height=100px").props(
                    "fit='scale-down'"
                )
